#include <string.h>

#include "gp32lib.h"

#include "gpmem.h"
#include "gpstdio.h"
#include "gpstdlib.h"

/*
References:
-----------

um_s3c2400_rev30.pdf (or um_s3c2400x_rev10.pdf) - available at Samsung site
(this includes the ARM cpu-manual)

UDA1330A.pdf                                    - available at Philips site
(this includes info about the soundchip and iis/l3 bus)

I2SBUS.pdf                                      - available at Philips site
(further general information about the iis bus)

test_24x_146u.zip                               - available at Samsung site
(invalueable sourcecode with a lot of lowlevel information inside)

remember the comments in this file won't cover anything very detailed that
is already contained in these files, so you might want to read these aswell.

Overview:
---------

+----------------------+                        +--------------------+
|                      |       (L3 Bus)         |                    |
| Samsung  PORT E b  9 |--------Commands------->| L3 Clock  Philips  |
| s3c2400         b 10 |                        |    Mode   uda1330a |
|                 b 11 |                        |    Data            |
|                      |                        |                    |
|           DMA 2      |-------Audio-Data------>| DAC                |
|                      |<------Audio-Data-------| ADC                |
|                      |       (IIS Bus)        |                    |
+----------------------+                        +--------------------+

The general process of playing sound on the Gp32 would be like:

1) stop any previously going sound-dma
2) setup the L3 bus
3) send commands to setup the uda1330a over L3 bus
4) setup DMA2
5) use DMA2 to send audio data to the uda1330a over IIS bus

you could probably play sound by direct cpu-writes aswell but i really
don't have a clue why you would want to do that ;)

*/

#define L3C             0x200   /* bit 9  */
#define L3M             0x400   /* bit 10 */
#define L3D             0x800   /* bit 11 */
#define L3_MASK         0xe00

#define L3DELAY			8   	/* delay for bus-transfer */

/*
   This routine sends an adress plus type of transfer to the L3 bus
*/
static void _WrL3Addr(unsigned char data)
{
	signed long i,j;

    rPEDAT = rPEDAT & ~L3_MASK;       			//L3D=L/L3M=L(in address mode)/L3C=L
    rPEDAT |= L3C;                    			//L3C=H
    for (j = 0; j < L3DELAY; j ++);           	//tsu(L3) > 190ns

    // PD[8:6]=L3D:L3C:L3M
    for (i = 0; i < 8; i ++)					//LSB first
    {
		if (data & 0x1)	//if data's LSB is 'H'
		{
			rPEDAT &= ~L3C;           			//L3C=L
			rPEDAT |= L3D;            			//L3D=H
			for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
			rPEDAT |= L3C;            			//L3C=H
			rPEDAT |= L3D;            			//L3D=H
			for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
		}
		else		//if data's LSB is 'L'
		{
            rPEDAT &= ~L3C;           			//L3C=L
            rPEDAT &= ~L3D;          			//L3D=L
            for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
            rPEDAT |= L3C;            			//L3C=H
            rPEDAT &= ~L3D;           			//L3D=L
            for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
		}
		data >>= 1;
    }

    //L3M=H,L3C=H
    rPEDAT = (rPEDAT & ~L3_MASK) | (L3C|L3M);
}

/*
   This routine sends a data word to the L3 bus
*/
static void _WrL3Data(unsigned char data,int halt)
{
	signed long i,j;

    if (halt)
    {

        rPEDAT |= L3C;                			//L3C=H(while tstp, L3 interface halt condition)
        for (j = 0; j < L3DELAY; j ++);       	//tstp(L3) > 190ns
        rPEDAT &= (~L3M);
        for (j = 0; j < L3DELAY; j ++);
        rPEDAT |= L3M;
    }

    rPEDAT = (rPEDAT & ~L3_MASK) | (L3C | L3M); //L3M=H(in data transfer mode)
    for (j = 0; j < L3DELAY; j ++);             //tsu(L3)D > 190ns

    //PD[8:6]=L3D:L3C:L3M
    for (i = 0; i < 8; i ++)
    {
        if (data & 0x1)	//if data's LSB is 'H'
        {
            rPEDAT &= ~L3C;           			//L3C=L
            rPEDAT |= L3D;            			//L3D=H
            for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
            rPEDAT |= (L3C|L3D);      			//L3C=H,L3D=H
            for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
        }
        else		//if data's LSB is 'L'
        {
            rPEDAT &= ~L3C;           			//L3C=L
            rPEDAT &= ~L3D;           			//L3D=L
            for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
            rPEDAT |= L3C;            			//L3C=H
            rPEDAT &= ~L3D;           			//L3D=L
            for (j = 0; j < L3DELAY; j ++);   	//tcy(L3) > 500ns
        }
        data >>= 1;	//for check next bit
    }

    rPEDAT = (rPEDAT & ~L3_MASK) | L3C;
    for (j = 0; j < L3DELAY; j ++);
    rPEDAT |= L3M;

}

/*
	This routine initializes the i/o port used for L3-bus and
	then configures the UDA1330
*/
static void InitUDA1330(void)
{
	/*
    //***** I/O Port E Initialize ******

    // port reconfiguration :
    //        PORTE 9,10,11 --> output
    //        pull-up disable
    //        L3_MODE, L3_CLK High

    rPEDAT = (rPEDAT & ~0xe00) | (L3M | L3C);
    rPEUP |= 0xe00;
    rPECON = (rPECON & (~(0x3f << 18))) | (0x15 << 18);

    //***** send commands via L3 Interface
    // Data bits 7 to 2 represent a 6-bit device address where bit 7 is the MSB.
    // The address of the UDA1330ATS is 000101 (bit 7 to bit 2).
    //*****

    // status type transfer , data value - clock=512fs, MSB format
    _WrL3Addr(0x14 + 2); 		// STATUS (0) (000101xx+10)
    _WrL3Data((0 << 4) +  		//  00,  : 512,384,256fs         (SC : System Clock Freq)
              (4 << 1),  		// 000,  : iis,lsb16,lsb18,lsb20,msb
              1);

    // data type transfer , data value - full volume
    _WrL3Addr(0x14 + 0);     	// DATA0 (000101xx+00)
    _WrL3Data(0x3f,           	// volume
              1);

    // data type transfer , data value - de-emphasis, no muting
    _WrL3Addr(0x14 + 0);     	// DATA0 (000101xx+00)
    _WrL3Data(0x80 +          	// select de-emhasis/mute
              (2 << 3),         // de-emp 44khz
              1);
	*/

	// set up clock speed to 256fs, iis format
	_WrL3Addr(0x14+2); //STATUS (0) (000101xx+10)
	_WrL3Data((2 << 4) + // 256fs
	          (0 << 1) // iis format
	          ,0);

	// set volume to 0 (seems that max volume is 0 or 1 not 0x3f *smile*)
	_WrL3Addr(0x14 + 0); // DATA0 (000101xx+00)
	_WrL3Data(0x0 // volume 0
	          ,0);
}
static void InitIIS(void)
{
    /****** IIS Initialize ******/
    // prescaler for 44khz
	// 67.8, 3
	rIISPSR = (2 << 5) + 2;	// Prescaler_A/B=4 for 11.2896MHz
	// 133, 3
	//rIISPSR = (5 << 5) + 5;	// Prescaler_A/B=5 for 5.6448MHz

	// prescaler for 22khz
	//rIISPSR = (9 << 5) + 9;	// Prescaler_A/B=9 for 5.6448MHz
	//rIISPSR = (5 << 5) + 5;	// Prescaler_A/B=5 for 5.6448MHz

    rIISCON = // 8,7,6 readonly
            (1 << 5) +  		// dma transmit request enable
            (0 << 4) +  		// dma recieve request disable
            (0 << 3) +  		// Transmit LR-Clock idle state
            (1 << 2) +  		// Recieve LR-Clock idle state
            (1 << 1) +  		// prescaler enabled
            (0 << 0);  			// iis disabled

    rIISMOD = (0 << 8) +      	// master mode
              (2 << 6) +      	// transmit mode
              (1 << 5) +      	// lowbyte=left channel
              (0 << 4) +      	// iis compatible samples /Changed this to IIS
              (1 << 3) +      	// 16 bit per channel
              (0 << 2) +      	// CDCLK=256fs
              (1 << 0);       	// serial bitclock 32fs


    rIISFIFCON = (1 << 11) +  	// transmit fifo mode = dma
                 (0 << 10) +  	// recieve fifo mode = normal
                 (1 << 9) +   	// transmit fifo enable
                 (0 << 8);    	// recieve fifo disable
                // 7-4 Transmit Fifo Data Count (readonly)
                // 3-0 Recieve Fifo Data Count (readonly)

    /****** IIS Tx Start ******/
    rIISCON |= 0x1;  // iis enable
}
static void CloseIIS(void)
{
    /****** IIS Tx Stop ******/
    rIISCON     = 0x0;		// IIS stop
    rDMASKTRIG2 = (1 << 2);	// DMA2 stop
    rIISFIFCON  = 0x0;		// for FIFO flush
}

void PlayPcm(unsigned char* src, int size, int rpt)
{
    rDIDST2 = (1 << 30) +     	// destination on peripherial bus
              (1 << 29) +       // fixed adress
              ((int) IISFIF); 	// IIS FIFO txd
    rDISRC2 = (0 << 30) +       // source from system bus
              (0 << 29) +       // auto-increment
              (int) (src);    	// buffer adress
    rDCON2 = (1 << 30) +       	// handshake mode
             (0 << 29) +       	// DREQ and DACK are synchronized to PCLK (APB clock)
             (1 << 28) +       	// generate irq when transfer is done
             (0 << 27) +       	// unit transfer
             (0 << 26) +       	// single service
             (0 << 24) +       	// dma-req.source=I2SSDO
             (1 << 23) +       	// (H/W request mode)
             (rpt << 22) +     	// auto reload on/off
             (1 << 20) +       	// data size, byte,hword,word
             (size / 2);      	// transfer size (hwords)

    rDMASKTRIG2 = (0 << 2) + (1 << 1) + 0;  /* no-stop, DMA2 channel on, no-sw trigger */
}

BOOL PcmInit(void)
{
    /* stop dma2 */
    rDMASKTRIG2 = (1 << 2) + (0 << 1) + 0;  /* stop, DMA2 channel off, no-sw trigger */

    /* initialize uda1330a by sending appropriate commands over L3 bus */
    InitUDA1330();

    /* setup IIS bus */
    InitIIS();

    return TRUE;
}
void PcmClose(void)
{
	CloseIIS();
}

/*
Since the Table of Adresses for Interupt-Service-Routines appearently is not
mapped to a fixed memory location, the bios provides two system-calls (aka swi)
for manipulating it.
*/
void swi_install_irq(unsigned long nr,void *ptr);
void swi_uninstall_irq(unsigned long nr);

/*
If you want to modify the settings of an interupt, you are supposed to disable
any interupts before and enable them again after like mentioned below:
*/
void ARMEnableInterrupt(void);
void ARMDisableInterrupt(void);


///////////////////////////////////////////////////////////////////////////////
DWORD	dwPcmChunkOffset, dwPcmChunkLength;
LPBYTE	pPcmChunkBuffer;

static LPBYTE pPcmChunkOffset;

void PlayPcmNextChunk(void)
{
	static DWORD dwLength = CHUNK_LENGTH;

	// Play next chunk
	PlayPcm(pPcmChunkOffset, dwLength, 1);

	dwPcmChunkOffset += dwLength;
	if (dwPcmChunkOffset > (dwPcmChunkLength - 1))
	{
		dwLength = CHUNK_LENGTH;

		pPcmChunkOffset   = pPcmChunkBuffer;
		dwPcmChunkOffset -= dwPcmChunkLength;

		return;
	}

	dwLength = dwPcmChunkLength - dwPcmChunkOffset;
	if (dwLength > CHUNK_LENGTH)
		dwLength = CHUNK_LENGTH;

	pPcmChunkOffset += dwLength;
}

//static void DMA2_ISR(void) __attribute__ ((interrupt ("IRQ")));
static void __irq DMA2_ISR(void);
static void DMA2_ISR(void)
{
	PlayPcmNextChunk();
}

/*
        Last not least, install and activate your own interupt routine like this:
*/

#define nISR_DMA2       0x13
#define BIT_DMA2	(0x1<<19)

/* for other IRQ numbers/bits refer to CPU-Manual */
void DMA2_ISR_Install(void)
{
    /* set vector for DMA2_DONE interupt

       this systemcall copies the said vector to the table used by the interupt
       handler in the bios.

    */
    rINTMOD &= ~(BIT_DMA2);
    rINTMSK &= ~(BIT_DMA2);

    swi_install_irq(nISR_DMA2, (void*) &DMA2_ISR);

    /* set irq regs for DMA2_DONE interupt */
    rSRCPND  = BIT_DMA2;
    rINTPND  = BIT_DMA2;
    rINTMOD &= ~(BIT_DMA2);
    rINTMSK &= ~(BIT_DMA2);
}
void DMA2_ISR_Uninstall(void)
{
    swi_uninstall_irq(nISR_DMA2);
}

BOOL CreatePcmChunk(void)
{
    ///////////////////////////////////////////////////////////////////////////
	dwPcmChunkOffset = 0;
	dwPcmChunkLength = CHUNK_LENGTH * 4; // 16k

	pPcmChunkBuffer = (LPBYTE) gm_malloc(dwPcmChunkLength);
	if (!pPcmChunkBuffer)
		return FALSE;

	memset(pPcmChunkBuffer, 0, dwPcmChunkLength);

	pPcmChunkOffset = pPcmChunkBuffer;
}
void DestroyPcmChunk(void)
{
	if (pPcmChunkBuffer)
	{
 		gm_free(pPcmChunkBuffer);
 		pPcmChunkBuffer = NULL;
 	}
}

///////////////////////////////////////////////////////////////////////////////
void SetClockSpeed(int nClockSpeed)
{
    // prescaler for 44khz
	switch (nClockSpeed)
	{
	case CPU_CLOCK40:
		// 67.8, 3
		GpClockSpeedChange (40000000, 0x48013, 1);
		break;

	case CPU_CLOCK67:
		// 67.8, 3
		GpClockSpeedChange(67800000,  0x69032, 3);
		break;

	case CPU_CLOCK133:
		// 133, 3
		GpClockSpeedChange(132000000,  0x3a011, 3);
		break;
	}
}
